// filetree.hh
// file access implemented as a binary tree
// Kevin Puckett Albrecht
// Rev 1
// 23 March 2000 - 2 May 2001

#ifndef FILETREE_HH
#define FILETREE_HH

/////////////////////////
// DEFINES //////////////
/////////////////////////

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <fstream.h>

// Local //
#include "error.hh"
#include "common.hh"

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

class FileTree; // declare for use in FileVertex

class FileVertex
{
  public:
    fstream filehandle;
    bool bInput;        // true==input, false==output
    bool bOpen;         // true==open, false==closed

    FileVertex (char *asFilename);
    ~FileVertex ();

    FileVertex *GetLeft () const {return mpLeft;}
    FileVertex *GetRight () const {return mpRight;}

  private:
    FileVertex *mpLeft, *mpRight; // left and right subtrees
    char *msFilename;             // symbol string

    friend class FileTree;
};

//// Constructor ////
FileVertex::FileVertex (char *asFilename)
{
  mpLeft = mpRight = NULL;

  msFilename = new char[strlen(asFilename) + 1];
  k_copy (msFilename, asFilename);

  bInput = true;
  bOpen = true;
}

//// Destructor ////
FileVertex::~FileVertex ()
{
  // First the subtrees //
  delete mpLeft;
  delete mpRight;

  // Then delete this vertex's components //
  k_del (msFilename);
}

class FileTree
{
  public:
    FileTree ();
    ~FileTree ();

    FileVertex *Search (char *asFilename) const;
    FileVertex *EnterNew (char *asFilename);

    FileVertex *GetRoot () const {return mpRoot;}
    int GetVertices () const {return mnVertices;}

  private:
    FileVertex *mpRoot; // ptr to binary tree root
    short mnVertices; // vertex counter
};


//// Constructor ////
FileTree::FileTree ()
{
  mpRoot = NULL;
  mnVertices = 0;
}

//// Destructor ////
FileTree::~FileTree ()
{
  delete mpRoot;
}

//// Search ////
//// Returns pointer to the vertex, or NULL if not found
////////////////////////////////////////////////////////
FileVertex *FileTree::Search (char *asFilename) const
{
  FileVertex *pNode = mpRoot;
  int nCompare;

  // Loop to search the table //
  while (pNode)
  {
    nCompare = strcmp (asFilename, pNode->msFilename); // compare names

    // Found //
    if (nCompare == 0)
    {
      break;
    }

    // Not found; search left or right subtree //
    pNode = nCompare < 0 ? pNode->mpLeft : pNode->mpRight;
  }

  return pNode;  // ptr to node, or NULL if not found
}

//// EnterNew ////
//// If the vertex is found return NULL; Otherwise, create it.
//// Returns pointer to the vertex.
/////////////////////////////////////////////////////////////////
FileVertex *FileTree::EnterNew (char *asFilename)
{
  FileVertex *pNode; // ptr to node
  FileVertex **ppNode = &mpRoot; // ptr to ptr to node

  // Loop to search tree for insertion point //
  while ((pNode = *ppNode) != NULL)
  {
    // bug fix ?? //
	cout << "<" << pNode->msFilename << ">";
    int nCompare = -1;
    if (pNode) nCompare = strcmp (asFilename, pNode->msFilename);

    // Found //
    if (nCompare == 0)
    {
      // Vertex exists, but was closed; reopen; return address //
      if (pNode->bOpen == false)
      {
        pNode->bOpen = true;
        pNode->bInput = true;
        return pNode;
      }
      // Vertex exists already //
      else
      {
        return NULL;
      }
    }

    // Not found; search left or right subtree //
    ppNode = nCompare < 0 ? &(pNode->mpLeft) : &(pNode->mpRight);
  }

  // Create and insert a new vertex //
  pNode = new FileVertex (asFilename); // create a new vertex
  mnVertices++;                        // increment vertex counter
  *ppNode = pNode;                     // insert it
  return pNode;                        // return it
}

#endif