// icode.hh
// intermediate code class
// Kevin Puckett Albrecht
// Rev 4
// 31 March 2000 - 27 June 2000

#ifndef ICODE_HH
#define ICODE_HH

/////////////////////////
// DEFINES //////////////
/////////////////////////

// I/O modes //
#define ICODE_INPUT 0
#define ICODE_OUTPUT 1

#define ICODE_SEGMENT_SIZE 4096 // 2^12

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <memory.h> // memcpy, etc.

// Local //
#include "common.hh"
#include "tree.hh"
#include "error.hh"

#include "token.hh"
#include "eoftoken.hh"
#include "eoltoken.hh"
#include "errortoken.hh"
#include "numbertoken.hh"
#include "specialtoken.hh"
#include "stringtoken.hh"
#include "wordtoken.hh"

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

class ICode
{
  public:
    ICode () { mpCode = mpPos = new char[ICODE_SEGMENT_SIZE]; }
    ICode (const ICode &licode);

    ~ICode ();

    int GetCurrentPos () const { return mpPos - mpCode; }
    TreeNode *GetNode () const { return mpNode; }

    void GoTo (int anPos) { mpPos = mpCode + anPos; }
    void Reset () { mpPos = mpCode; }

    void Put (TokenCode aCode);
    void Put (TreeNode *apNode);
    void PutNumber (short an);
    void PutLineMarker ();

    Token *Get ();

  private:
    char *mpCode; // beginning of code segment
    char *mpPos;  // current position in code segment
    TreeNode *mpNode; // extracted tree node

    void CheckBounds (int anSize);
    TreeNode *GetTreeNode ();

    WordToken    wordToken;
    NumberToken  numberToken;
    StringToken  stringToken;
    SpecialToken specialToken;
    EOLToken     eolToken;
    EOFToken     eofToken;
    ErrorToken   errorToken;
};

//// Copy Constructor ////
//////////////////////////
ICode::ICode (const ICode &licode)
{
  int nLen = (int) (licode.mpPos - licode.mpCode);

  // Copy icode //
  mpCode = mpPos = new char[nLen];
  memcpy (mpCode, licode.mpCode, nLen);
}

//// Destructor ////
////////////////////
ICode::~ICode ()
{
  k_del (mpCode);
}

//// Put ////
//// Append an token to the icode.
//////////////////////////////////
void ICode::Put (TokenCode aCode)
{
  //removed// if (errorCount > 0) return;
    
  char cCode = aCode;
  CheckBounds (sizeof (char));
  memcpy ((void *) mpPos, (const void *) &cCode, sizeof (char));
  mpPos += sizeof (char);
}

//// Put ////
//// Append a tree node to the icode.
/////////////////////////////////////
void ICode::Put (TreeNode *apNode)
{
  //removed// if (errorCount > 0) return;

  short xTree = apNode->GetTreeIndex ();
  short xNode = apNode->GetNodeIndex ();

  CheckBounds (2 * sizeof (short));
  memcpy ((void *) mpPos, (const void *) &xTree, sizeof (short));
  memcpy ((void *) (mpPos + sizeof (short)), (const void *) &xNode, sizeof (short));
  mpPos += 2 * sizeof (short);
}

//// PutNumber ////
//// Append an integer to the icode.
////////////////////////////////////
void ICode::PutNumber (short an)
{
  char cCode = codeShortMarker;

  CheckBounds (sizeof (char) + sizeof (short));
  memcpy ((void *) mpPos, (const void *) &cCode, sizeof (char));
  mpPos += sizeof (char);
  memcpy ((void *) mpPos, (const void *) &an, sizeof (short));
  mpPos += sizeof (short);
}

//// PutLineMarker ////
//// Insert a line marker into the intermediate code just
//// ...before the last appended token code.
/////////////////////////////////////////////////////////
void ICode::PutLineMarker ()
{
  //removed// if (errorCount > 0) return;

  // Remember the last apended token code //
  char cLastCode;
  mpPos -= sizeof (char);
  memcpy ((void *) &cLastCode, (const void *) mpPos, sizeof (char));

  // Insert a line marker code followed by the current line number //
  char cCode = codeLineMarker;
  short nNum = gpCommon->nCurrentLine;
  CheckBounds (sizeof (char) + sizeof (short));
  memcpy ((void *) mpPos, (const void *) &cCode, sizeof (char));
  mpPos += sizeof (char);
  memcpy ((void *) mpPos, (const void *) &nNum, sizeof (short));
  mpPos += sizeof (short);
  
  // Re-append the last token code //
  memcpy ((void *) mpPos, (const void *) &cLastCode, sizeof (char));
  mpPos += sizeof (char);
}

//// Get ////
//// Extract a token from the icode.
////////////////////////////////////
Token *ICode::Get ()
{
  Token *pToken; // token to return
  char cCode; // token code read from the file
  TokenCode code;
  short shortMarkerValue = 0; // value from PutNumber()

  do
  {
    memcpy ((void *) &cCode, (const void *) mpPos, sizeof (char));
    mpPos += sizeof (char);
    code = (TokenCode) cCode;

    if (code == codeLineMarker)
    {
      short nNum;
      memcpy ((void *) &nNum, (const void *) mpPos, sizeof (short));
      gpCommon->nCurrentLine = nNum;
      mpPos += sizeof (short);
    }
    else if (code == codeShortMarker)
    {
      memcpy ((void *) &shortMarkerValue, (const void *) mpPos, sizeof (short));
      mpPos += sizeof (short);
    }
  } while (code == codeLineMarker);

  // Determine the token class //
  switch (code)
  {
    case codeIdentifier:
    {
      pToken = &wordToken;
      pToken->code = codeIdentifier;
      break;
    }
    case codeNumber:
      pToken = &numberToken;
      break;
    case codeString:
      pToken = &stringToken;
      break;
    case codeEOL:
      pToken = &eolToken;
      break;
    case codeEOF:
      pToken = &eofToken;
      break;
    case codeShortMarker:
      pToken = &specialToken;
      pToken->code = code;
      break;
    default:
    {
      if (code < codeCOLON) // reserved word
      {
        pToken       = &wordToken;
        pToken->code = code;
      }
      else // operator
      {
        pToken       = &specialToken;
        pToken->code = code;
      }
      break;
    }
  }

  // Extract and set the token string //
  switch (code)
  {
    case codeIdentifier:
    case codeNumber:
    case codeString:
    {
      mpNode = GetTreeNode ();
      strcpy (pToken->buffer, mpNode->GetString ());
      break;
    }
    case codeShortMarker:
    {
      mpNode = NULL;
      strcpy (pToken->buffer, " ");
      pToken->marker = shortMarkerValue;
      break;
    }
    default: // codeEOL, codeEOF, reserved word, or operator
    {
      mpNode = NULL;
      k_copy (pToken->buffer, gvsTokenStrings[int(cCode)]);
      break;
    }
  }

  return pToken;
}

//// CheckBounds ////
//// Guard against code segment overflow.
/////////////////////////////////////////
void ICode::CheckBounds (int anSize)
{
  if ((mpPos + anSize) >= &mpCode[ICODE_SEGMENT_SIZE])
  {
    gpErrorR->Show (EFF_TERMINATE, ER_CODE_SEGMENT_OVERFLOW);
  }
}

//// GetTreeNode ////
//// Extract a tree node from the icode.
////////////////////////////////////////
TreeNode *ICode::GetTreeNode ()
{
  short xTree;
  short xNode;

  memcpy ((void *) &xTree, (const void *) mpPos, sizeof (short));
  memcpy ((void *) &xNode, (const void *) (mpPos + sizeof (short)), sizeof (short));

  mpPos += 2 * sizeof (short);

  return gvpTrees[xTree]->GetNode (xNode);
}

#endif
