// tree.hh
// symbol table implemented as a binary tree
// Kevin Puckett Albrecht
// Rev 8
// 23 March 2000 - 02 May 2001
//
// must define: GLOBAL_VALUE_STR_SIZE

#ifndef TREE_HH
#define TREE_HH

/////////////////////////
// DEFINES //////////////
/////////////////////////

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //
#include "inbuffer.hh"
#include "common.hh"
#include "value.hh"

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

class Tree; // declare for use in TreeNode

class TreeNode
{
  public:
    Type type; // the type in the node
    bool bConst; // true if the node's contents are constant
    bool bDefined; // true if type == tValue and execution has passed 
                   //  it's definition (only used in Exec)
    bool bInDeclaration; // true if the variable is currently being
                         //  declared (only used in Parser)
    int pos; // the position in the string, if type == tValue
    short marker; // the marker in the node, if type == tMarker

    TreeNode (char *apString, Type aType = tDummy);
    ~TreeNode ();

    void Convert (TreeNode **avpNodes);

    TreeNode *GetLeft () const {return mLeft;}
    TreeNode *GetRight () const {return mRight;}
    char *GetString () const {return mpString;}
    short GetTreeIndex () const {return mxTree;}
    short GetNodeIndex () const {return mxNode;}

    char *GetValue () const {return k_clone(msValue);}

    void SetValue (char *asValue)
    {
      // The counter is reset only when the length of the string is decreased //
      if (strlen (msValue) > strlen (asValue)) pos = 0;

      k_del (msValue);
      msValue = k_clone (asValue);
    }
    void SetValueD (char *asValue)
    {
      // The counter is reset only when the length of the string is decreased //
      if (strlen (msValue) > strlen (asValue)) pos = 0;

      k_del (msValue);
      msValue = k_clone (asValue);
      k_del (asValue);
    }

    void SetContent (Type aType, bool abConst, char *asValue, short anMarker = 0)
    {
      type = aType;
      bConst = abConst;

      // The counter is reset only when the length of the string is decreased //
      if (strlen (msValue) > strlen (asValue)) pos = 0;

      k_del (msValue);
      msValue = k_clone (asValue);

      marker = anMarker;
    }

  private:
    TreeNode *mLeft, *mRight; // left and right subtrees
    char *mpString;           // symbol string
    short mxTree;             // tree index
    short mxNode;             // node index
    char *msValue;            // the value in the node, if type == tValue

    friend class Tree;
};

//// Constructor ////
TreeNode::TreeNode (char *apString, Type aType)
{
  type = aType;
  bConst = false;
  bDefined = false;
  bInDeclaration = false;
  msValue = new char[GLOBAL_VALUE_STR_SIZE]; k_copy (msValue, "");
  marker = 0;

  pos = 0;

  mLeft = mRight = NULL;
  mxNode = 0;

  // Allocate and copy the symbol string //
  mpString = new char[strlen(apString) + 1];
  k_copy (mpString, apString);
}

//// Destructor ////
TreeNode::~TreeNode ()
{
  // First the subtrees //
  delete mLeft;
  delete mRight;

  // Then delete this node's components //
  k_del (msValue);
  k_del (mpString);
}

//// Convert ////
//// "Flatten" the node.
////////////////////////
void TreeNode::Convert (TreeNode **avpNodes)
{
  // Convert the left //
  if (mLeft) mLeft->Convert (avpNodes);

  avpNodes[mxNode] = this;

  // Convert the right //
  if (mRight) mRight->Convert (avpNodes);
}

class Tree
{
  private:
    TreeNode *mRoot; // ptr to binary tree root
    TreeNode **mvpNodes; // ptr to vector of nodes
    short mnNodes; // node counter
    short mxTree; // symbol table index
    Tree *mpNextTree;

  public:
    Tree ();
    ~Tree ();

    TreeNode *Search (char *apString) const;
    TreeNode *Enter (char *apString, Type aType = tDummy);
    TreeNode *EnterNew (char *apString, Type aType = tDummy);
    void Convert (Tree **avpTrees);

    TreeNode *GetRoot () const {return mRoot;}
    TreeNode *GetNode (short axNode) const {return mvpNodes[axNode];}
    Tree *GetNextTree () const {return mpNextTree;}
    TreeNode **GetNodeVector () const {return mvpNodes;}
    int GetNodeCount () const {return mnNodes;}
};

//// Constructor ////
Tree::Tree ()
{
  extern Tree *gpTreeList;

  mRoot = NULL;
  mvpNodes = NULL;
  mnNodes = 0;
  mxTree = gpCommon->nTrees++;

  // Insert at the head of the tree list //
  mpNextTree = gpTreeList;
  gpTreeList = this;
}

//// Destructor ////
Tree::~Tree ()
{
  delete mRoot;
  delete [] mvpNodes;
}

//// Search ////
//// Returns pointer to the node, or NULL if not found
//////////////////////////////////////////////////////
TreeNode *Tree::Search (char *apString) const
{
  TreeNode *pNode = mRoot;  // ptr to symbol table node
  int nCompare;

  // Loop to search the table //
  while (pNode)
  {
    nCompare = strcmp (apString, pNode->mpString);  // compare names
    if (nCompare == 0) break;                       // found!

    // Not found; Search left or right subtree //
    pNode = nCompare < 0 ? pNode->mLeft : pNode->mRight;
  }

  return pNode;  // ptr to node, or NULL if not found
}

//// Enter ////
//// Returns pointer to the node; If not found, it will be created
//////////////////////////////////////////////////////////////////
TreeNode *Tree::Enter (char *apString, Type aType)
{
  TreeNode *pNode; // ptr to node
  TreeNode **ppNode = &mRoot; // ptr to ptr to node

  // Loop to search tree for insertion point //
  while ((pNode = *ppNode) != NULL)
  {
    int nCompare = strcmp (apString, pNode->mpString);  // compare strings
    if (nCompare == 0) return pNode;                    // found!

    // Not found; Search left or right subtree //
    ppNode = nCompare < 0 ? &(pNode->mLeft) : &(pNode->mRight);
  }

  // Create and insert a new node //
  pNode = new TreeNode (apString, aType); // create a new node
  pNode->mxTree = mxTree;                  // set tree index
  pNode->mxNode = mnNodes++;               // set node index
  *ppNode = pNode;                         // insert it
  return pNode;                            // return it
}

//// EnterNew ////
//// If the node is found, flag a error; Otherwise, create it.
//// Returns pointer to the node.
//////////////////////////////////////////////////////////////
TreeNode *Tree::EnterNew (char *apString, Type aType)
{
  TreeNode *pNode = Search (apString);

  if (!pNode) pNode = Enter (apString, aType);
  else gpErrorS->Show (EFF_RECOVER, ES_ALREADY_DECLARED, apString);

  return pNode;
}

//// Convert ////
//// "Flatten" the tree.
////////////////////////
void Tree::Convert (Tree **avpTrees)
{
  avpTrees[mxTree] = this;

  mvpNodes = new TreeNode *[mnNodes];

  // checking if mRoot exists fixes the problem of an empty file //
  if (mRoot) mRoot->Convert (mvpNodes);
}

#endif